/* Datei: hash3.c */
/* Funktionen zur Verwaltung von Hashtabellen mit direkter        */
/* Verkettung als Konfliktstrategie. Die Hashtabelle enthaelt nur */
/* Pointer auf die Elemente.                                      */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#define GRENZE       11
#define FUELLWERT    20
#define FREI         NULL
typedef unsigned int uint;
typedef struct element { uint schl;
                         char name [6];
                         struct element * nachfolger;
                       } ELEMENT;

ELEMENT * tabelle [GRENZE]; /* Nur Pointer auf Elemente           */
void einfuegen (ELEMENT, ELEMENT * []);
void suche_element (uint, ELEMENT * [], ELEMENT **, int *);
uint hash (uint);
static void init_tabelle (ELEMENT * []);
static void fuelle_tabelle (ELEMENT * []);
void ausgeben_tabelle (ELEMENT * []);

int main (void)
{
   uint schl;
   ELEMENT * el;
   char zeile [80];
   int i;
   ELEMENT * ptr, * next;

   init_tabelle (tabelle);
   fuelle_tabelle (tabelle);
   ausgeben_tabelle (tabelle);

   /* Test der Suchfunktion                                       */
   while (printf ("\nSuchschluessel = "), gets (zeile),
          sscanf (zeile, "%d", &schl) != 0)
   {
      suche_element (schl, tabelle, &el, &i);
      if (el != NULL)
         printf ("Schluessel = %d Name = %s index = %d\n", 
                 schl, tabelle[i]->name, i);
      else
         printf ("Schluessel %d nicht in der Hashtabelle\n", schl);
   }

   /* Loeschen der Tabellenelemente                               */
   for (i = 0; i < GRENZE; i++)
   {
      ptr = tabelle[i];
      while (ptr != NULL)
      {
         next = ptr->nachfolger;
         free (ptr);
         ptr = next;
      }
   }
   return 0;
}

static void init_tabelle (ELEMENT * tab [])
{  /* Alle Pointer der Tabellenelemente mit "FREI" markieren.     */
   int i;
   for (i = 0; i < GRENZE; i++) tab[i] = FREI;
}

uint hash (uint schl)
{  /* Einfache Hashfunktion                                       */
   return (schl % GRENZE);
}

static void fuelle_tabelle (ELEMENT ** tab)
{  /* Generiert "FUELLWERT" Zufallszahlen zwischen 1 und 32767    */
   /* und "Zufallsnamen" von "A" .. "Z"                           */
   int i;
   ELEMENT y;

   srand(0) ;
   for (i = 0; i < FUELLWERT; i++) 
   {  /* Zufaellige Schluessel erzeugen. Hier wird willkuerlich   */
      /* ein Name erzeugt, der nur einen Buchstaben (A..Z) hat.   */
      y.schl = 1 + rand() % 32767;
      y.name[0] = 'A' + i % 26;
      y.name[1] = '\0';
      einfuegen (y, tab);
   }
}

void einfuegen (ELEMENT x , ELEMENT * tab[])
{  /* Fuegt Element x mit Schluessel "schl" in die Hashtab. ein   */
   int ketten_laenge = 0;
   int ind;
   ELEMENT * el;

   ind = hash (x.schl);
   x.nachfolger = NULL;

   if ((tab[ind] == FREI))
   {
      if ((tab[ind] = (ELEMENT *)malloc (sizeof (ELEMENT))) != NULL)
         memcpy (tab[ind], &x, sizeof (x));
      else
         exit(1);
   }
   else
   {  /* Ende der Kette suchen                                    */
      for (el = tab[ind]; el->nachfolger != NULL;
           el = el->nachfolger)
         ketten_laenge++;
      /* Neues Element anhaengen */
      if ((el->nachfolger = 
           (ELEMENT *)malloc (sizeof (ELEMENT))) != NULL)
         memcpy (el->nachfolger, &x, sizeof (x));
      else
         exit(1);
   }
}

void suche_element (uint schl, ELEMENT * tab [], ELEMENT ** x,
                    int * pos)
{  /* Sucht Element el mit Schluessel "schl" in Hashtabelle "tab".*/
   /* Wenn gefunden :   return (x , ind) ,                        */
   /*         Sonst :   return (x = NULL , ind = ??)              */
   uint ind;
   ELEMENT *el;

   ind = hash (schl);
   *pos = ind;

   /* Element in der Kette suchen                                 */
   for (el = tab[ind]; el != NULL; el = el->nachfolger)
       if (el->schl == schl)
       {
          *x = el;
          return;
       }
   *x = NULL ;
}

void ausgeben_tabelle (ELEMENT * tab [])
{
   int i;
   ELEMENT * ptr;
   printf ("\nIndex | Schluessel | Name");
   printf ("\n------|------------|-----");
   for(i = 0; i < GRENZE;i++)
   {
      if (tab[i] != FREI)
      {
         ptr = tab[i];
         while (ptr != NULL)
         {
            printf("\n  %2d  |    %5d   |  %s", i, 
                    ptr->schl, ptr->name);
            ptr = ptr->nachfolger;
         }
      }
   }
}

